@@ -3,8 +3,9 @@ from .ninjax import __version__
 # Basics
 from .ninjax import Module
 from .ninjax import Variable
-from .ninjax import pure
+from .ninjax import Tree
 from .ninjax import init
+from .ninjax import pure
 from .ninjax import seed
 
 # Transforms
@@ -19,6 +20,8 @@ from .ninjax import creating
 from .ninjax import scope
 
 # Integrations
+from .ninjax import flatten
+from .ninjax import unflatten
 from .ninjax import FromHaiku
 from .ninjax import FromFlax
 from .ninjax import FromOptax
