@@ -7,7 +7,14 @@ import threading
 import jax
 import jax.numpy as jnp
 
-__version__ = '2.3.1'
+__version__ = '3.1.0'
+
+
+def add_note(e, note):
+  if hasattr(e, 'add_note'):
+    e.add_note(note)
+  else:
+    print(note)
 
 
 ###############################################################################
@@ -113,14 +120,14 @@ def pure(fun, nested=False):
       ignore = ignore if ignore is not None else False
     if not isinstance(state, dict):
       raise ValueError('Must provide a dict as state.')
+    name = getattr(fun, '__name__', str(fun))
     if context and (not nested):
       raise RuntimeError(
-          f'You are trying to call pure {fun.__name__}() inside pure '
+          f'You are trying to call pure {name}() inside pure '
           f'{context.name}(). Is that intentional? If you want to nest pure '
           f'functions, use pure(..., nested=True) for the inner function.')
     before = context
     try:
-      name = fun.__name__
       context = Context(
           state.copy(), seed, create, modify, ignore, [], name)
       CONTEXT[threading.get_ident()] = context
@@ -197,36 +204,41 @@ def creating():
 
 
 @jax.named_scope('grad')
-def grad(fun, keys, has_aux=False):
+def grad(fun, targets, has_aux=False):
   """Compute the gradient of an impure function with respect to the specified
   state entries or modules. The transformed function returns a tuple containing
   the computed value, selected state entries, their gradients, and if
   applicable auxiliary outputs of the function."""
-  keys = keys if hasattr(keys, '__len__') else (keys,)
+  targets = targets if hasattr(targets, '__len__') else (targets,)
   if not has_aux:
-    fun = lambda *args, _fun=fun, **kwargs: (_fun(*args, *kwargs), {})
+    fun = lambda *args, _fun=fun, **kwargs: (_fun(*args, **kwargs), {})
   fun = pure(fun, nested=True)
 
   def wrapper(*args, **kwargs):
     accessed, modified = _prerun(fun, *args, **kwargs)
 
     strs = []
-    for key in keys:
-      if isinstance(key, Module):
-        matches = key.find()
-      if isinstance(key, str):
-        pattern = re.compile(f'^{key}(/.*|$)')
+    for target in targets:
+      if isinstance(target, Module):
+        prefix = target.path + '/'
+        matches = {k: v for k, v in context().items() if k.startswith(prefix)}
+      if isinstance(target, str):
+        pattern = re.compile(f'^{target}(/.*|$)')
         matches = [k for k in context() if pattern.match(k)]
       if not matches:
+        existing = ', '.join(context().keys())
         raise KeyError(
-            f"Gradient key '{key}' did not match any state entries. "
-            'List existing entries using print(nj.context().keys()).')
+            f"Gradient target '{target}' did not match any state entries. " +
+            f'Existing state entries: {existing}')
       strs += matches
     existing = context().keys()
     assert all(key in existing for key in strs), (strs, existing)
     x1 = {k: v for k, v in context().items() if k in strs}
     x2 = {k: v for k, v in context().items() if k not in strs}
-    assert x1
+    if not x1:
+      raise ValueError(
+          'No inputs to differentiate with respect to. ' +
+          f"User provided targets: '{targets}'")
 
     for key in x1.keys():
       if key not in accessed:
@@ -283,13 +295,13 @@ def cond(pred, true_fun, false_fun, *operands):
 
 
 @jax.named_scope('scan')
-def scan(fun, carry, xs, reverse=False, unroll=1, axis=0):
+def scan(fun, carry, xs, length=None, reverse=False, unroll=1, axis=0):
   if axis:
-    xs = jax.tree_util.tree_map(lambda x: x.swapaxes(0, axis), xs)
+    xs = jax.tree.map(lambda x: x.swapaxes(0, axis), xs)
 
   fun = pure(fun, nested=True)
   accessed, modified = _prerun(
-      fun, carry, jax.tree_util.tree_map(lambda x: x[0], xs))
+      fun, carry, jax.tree.map(lambda x: x[0], xs))
 
   changing = {k: v for k, v in context().items() if k in modified}
   unchanging = {
@@ -304,7 +316,8 @@ def scan(fun, carry, xs, reverse=False, unroll=1, axis=0):
     changing = {k: v for k, v in state.items() if k in modified}
     return (carry, changing), y
 
-  length = len(jax.tree_util.tree_leaves(xs)[0])
+  if length is None:
+    length = len(jax.tree.leaves(xs)[0])
   seeds = seed(length, True)
   (carry, changes), ys = jax.lax.scan(
       inner, (carry, changing), (xs, seeds), length, reverse, unroll)
@@ -313,7 +326,7 @@ def scan(fun, carry, xs, reverse=False, unroll=1, axis=0):
     context().update(changes)
 
   if axis:
-    ys = jax.tree_util.tree_map(lambda y: y.swapaxes(0, axis), ys)
+    ys = jax.tree.map(lambda y: y.swapaxes(0, axis), ys)
   return carry, ys
 
 
@@ -348,6 +361,9 @@ def checkpoint(fun, **cp_kwargs):
 def _prerun(fun, *args, **kwargs):
   if not context().modify and not context().create:
     return set()
+  # Copy container structure so modifications inside the user function
+  # (e.g. popping from a dict) are not applied during prerun.
+  args, kwargs = jax.tree.map(lambda x: x, (args, kwargs))
   state, output, accessed, modified, created = fun(
       dict(context()), *args, ignore=True, track=True,
       seed=seed(None, True), **kwargs)
@@ -375,25 +391,34 @@ def scope(name, absolute=False):
         'Purify stateful functions with fn = pure(fn) before running them.')
   outside = SCOPE
   if absolute:
+    validate(name)
     SCOPE = name
   elif SCOPE == '':
     SCOPE = name
   else:
+    validate(name, single=True)
     SCOPE = outside + '/' + name
   try:
     yield SCOPE
   except Exception as e:
     if not hasattr(e, '_njscope'):
       e._njscope = SCOPE
-      if hasattr(e, 'add_note'):
-        e.add_note(f"This happened inside Ninjax scope '{SCOPE}'.")
-      else:
-        print(f"Exception happened inside Ninjax scope '{SCOPE}'.")
+      add_note(e, f"Exception happened inside Ninjax scope '{SCOPE}'.")
     raise
   finally:
     SCOPE = outside
 
 
+def validate(path, single=False):
+  names = path.split('/')
+  assert not single or len(names) == 1, (path, names)
+  for name in names:
+    assert name, ('Name cannot be empty', path, name)
+    assert '{' not in name, ('Did you forget to format a string?', path, name)
+    assert re.match(r'^[A-Za-z0-9_]+$', name), (
+        'Only letters, numbers, and underscores allowed in names', path, name)
+
+
 class ModuleMeta(type):
   """Meta class that creates a unique path for each module instance and wraps
   the methods and properties of the module to enter the name scope."""
@@ -414,15 +439,23 @@ class ModuleMeta(type):
       elif inspect.isfunction(value):
         method_names.append(key)
     cls = super(ModuleMeta, mcs).__new__(mcs, name, bases, clsdict)
-    cls.__field_defaults = {
+    for name, typ in cls.__annotations__.items():
+      try:
+        isinstance(0, typ)
+      except Exception:
+        raise ValueError(
+            f"Annotation '{typ}' for field '{key}' is not a valid type.")
+    cls.__defaults = {
         k: getattr(cls, k) for k, v in cls.__annotations__.items()
         if hasattr(cls, k)}
     for key, value in cls.__annotations__.items():
       setattr(cls, key, property(lambda self, key=key: self.__fields[key]))
-    for method_name in method_names:
-      method = getattr(cls, method_name)
+    for name in method_names:
+      if name in cls.__defaults:
+        continue
+      method = getattr(cls, name)
       method = _scope_method(method)
-      setattr(cls, method_name, method)
+      setattr(cls, name, method)
     return cls
 
   def __call__(cls, *args, name=None, **kwargs):
@@ -431,19 +464,16 @@ class ModuleMeta(type):
     if not isinstance(name, str):
       raise TypeError(
           "Please provide a module name via Module(..., name='example').")
-    if not re.match(r'^[A-Za-z0-9_]+$', name):
-      raise ValueError(
-          'Only letters, numbers, and underscores are allowed in scope names; '
-          f'got: {name}')
+    validate(name, single=True)
     fields = {}
     for key, typ in cls.__annotations__.items():
       if key in kwargs:
         value = kwargs.pop(key)
-      elif key in cls.__field_defaults:
-        value = cls.__field_defaults[key]
+      elif key in cls.__defaults:
+        value = cls.__defaults[key]
       else:
         raise TypeError(
-            f"Pass a keyword argument for field '{key}' or define a default.")
+            f"Pass a keyword arg for field '{key}' or define a default.")
       if typ is not None and not isinstance(value, typ):
         raise TypeError(
             f"Value '{value}' for field '{key}' is not of type "
@@ -455,9 +485,19 @@ class ModuleMeta(type):
       obj._path = path
     obj._submodules = {}
     init = _scope_method(cls.__init__)
-    init(obj, *args, **kwargs)
+    try:
+      init(obj, *args, **kwargs)
+    except TypeError as e:
+      if kwargs:
+        keys = ', '.join(sorted(kwargs.keys()))
+        add_note(e, f'Keyword arguments not matched to class fields: {keys}')
+      raise
     return obj
 
+  @property
+  def defaults(self):
+    return self.__defaults
+
 
 def _scope_method(method):
   @functools.wraps(method)
@@ -465,6 +505,7 @@ def _scope_method(method):
     with scope(self._path, absolute=True):
       with jax.named_scope(self._path.split('/')[-1]):
         return method(self, *args, **kwargs)
+  wrapper._method = method  # Debug info.
   return wrapper
 
 
@@ -485,72 +526,113 @@ class Module(object, metaclass=ModuleMeta):
     """The name of this module instance as a string."""
     return self._path.split('/')[-1]
 
-  def get(self, name, *args, **kwargs):
-    """Retrieve or create a state entry that belongs to this module."""
-    assert '{' not in name, 'Did you forget to format a string?'
+  @property
+  def values(self):
+    p = self.path + '/'
+    ctx = context()
+    # Read keys individually to mark them as accessed.
+    return {k.removeprefix(p): ctx[k] for k in ctx if k.startswith(p)}
+
+  def value(self, name, make, *args, **kwargs):
+    """Define and read a state entry in the scope of this module."""
+    validate(name)
     path = self.path + '/' + name
-    if name in self._submodules:
-      return self._submodules[name]
-    if path in context():
-      return context()[path]
-    ctor, *args = args
-    if 'name' in inspect.signature(ctor).parameters:
-      kwargs['name'] = name
-    value = ctor(*args, **kwargs)
-    # We support trees of arrays for easier integration with other libraries.
-    flat = jax.tree_util.tree_leaves(value)
-    if all(isinstance(x, jnp.ndarray) for x in flat):
+    if path not in context():
+      if callable(make):
+        value = make(*args, **kwargs)
+      else:
+        assert not args and not kwargs
+        value = make
       context()[path] = value
-      # Look up the value again to make sure we are registering it as an
-      # accessed key in the context.
-      return context()[path]
-    else:
-      self._submodules[name] = value
-      return value
+    # Look up the value again to register it as accessed.
+    return context()[path]
 
-  def put(self, *args):
-    """Update or create state entries that belong to this module. The arguments
-    are either string name and value of a single state entry or a dict holding
-    multiple state entries."""
-    if len(args) == 2:
-      name, value = args
-      self.put({self.path + '/' + name: value})
-      return value
-    assert len(args) == 1 and isinstance(args[0], dict)
-    mapping = args[0]
-    prefix = self.path + '/'
-    for key in mapping:
-      if not key.startswith(prefix):
-        raise KeyError(f'Key {key} does not belong to module {self.path}.')
-    context().update(mapping)
+  def read(self, name):
+    """Read a state entry in the scope of this module."""
+    validate(name)
+    return context()[self.path + '/' + name]
 
-  def find(self, pattern=r'.*', empty_ok=False):
-    """Find the state entries of this module, optionally filtered by regex."""
-    pattern = re.compile(pattern)
-    prefix = self.path + '/'
-    results = {}
-    for key, value in context().items():
-      if not key.startswith(prefix):
-        continue
-      if pattern.match(key[len(prefix):]):
-        results[key] = value
-    if not empty_ok and not results:
-      raise KeyError(f'Pattern {pattern} matched no state keys.')
-    return results
+  def write(self, name, value):
+    """Update the value of a state entry in the scope of this module."""
+    validate(name)
+    path = self.path + '/' + name
+    existing = context()[path]
+    value = jnp.asarray(value, dtype=existing.dtype)
+    assert existing.shape == value.shape, (existing.shape, value.shape)
+    context()[path] = value
+    # Return value without lookup because it was provided by the user and thus
+    # has to be available in the pure function already.
+    return value
+
+  def sub(self, name, make=None, *args, **kwargs):
+    """Define and retrieve a sub module."""
+    validate(name, single=True)
+    if name not in self._submodules:
+      assert make, 'Provide constructor for submodule that does not exist.'
+      assert SCOPE.startswith(self.path), 'Must be inside the parent module.'
+      module = make(*args, **kwargs, name=name)
+      self._submodules[name] = module
+    return self._submodules[name]
 
 
 class Variable(Module):
 
-  def __init__(self, ctor, *args, **kwargs):
-    self.ctor = ctor
-    self.args = args
-    self.kwargs = kwargs
+  def __init__(self, make, *args, **kwargs):
+    self.make = functools.partial(make, *args, **kwargs)
 
   def read(self):
-    return self.get('value', self.ctor, *self.args, **self.kwargs)
+    if not self.values:
+      self.value('value', self.make)
+    return super().read('value')
 
   def write(self, value):
-    return self.put('value', value)
+    if not self.values:
+      self.value('value', self.make)
+    return super().write('value', value)
+
+
+class Tree(Module):
+
+  def __init__(self, make, *args, **kwargs):
+    self.make = functools.partial(make, *args, **kwargs)
+
+  def read(self):
+    if not self.values:
+      mapping, self.treedef = flatten(self.make())
+      [self.value(k, v) for k, v in mapping.items()]
+    return unflatten(self.values, self.treedef)
+
+  def write(self, tree):
+    if not self.values:
+      mapping, self.treedef = flatten(self.make())
+      [self.value(k, v) for k, v in mapping.items()]
+    mapping, treedef = flatten(tree)
+    assert treedef == self.treedef, (self.treedef, treedef)
+    write = super().write
+    [write(k, v) for k, v in mapping.items()]
+    return jax.tree.map(lambda x: x, tree)
+
+
+def flatten(tree):
+  items, treedef = jax.tree_util.tree_flatten_with_path(tree)
+  paths, values = zip(*items)
+  tostr = lambda x: re.sub(
+      r'[^a-z0-9-_/]+', '_', str(x).lower()).replace('_/', '').replace('_', '')
+  strpaths = [[tostr(x) for x in path] for path in paths]
+  keys = ['/'.join(x for x in strpath if x) for strpath in strpaths]
+  if len(set(keys)) < len(keys):
+    raise ValueError(
+        'Cannot flatten PyTree to dict because paths are ambiguous '
+        'after converting them to string keys.\n'
+        'Paths: {paths}\nKeys: {keys}')
+  items = sorted(list(zip(keys, values)), key=lambda x: x[0])
+  return dict(items), treedef
+
+
+def unflatten(mapping, treedef):
+  items = sorted(list(mapping.items()), key=lambda x: x[0])
+  _, values = zip(*items) if items else ([], [])
+  return jax.tree.unflatten(treedef, values)
 
 
 ###############################################################################
@@ -558,41 +640,86 @@ class Variable(Module):
 ###############################################################################
 
 
-def FromFlax(ctor):
+def FromFlax(make, postinit=None):
+
   class FlaxModule(Module):
+
     def __init__(self, *args, **kwargs):
-      self.module = ctor(*args, **kwargs)
+      if callable(make):
+        self.module = make(*args, **kwargs)
+      else:
+        assert not args and not kwargs
+        self.module = make
+      self.treedef = None
+
     def __call__(self, *args, **kwargs):
-      seed_ = seed() if creating() else None
-      state = self.get('flax', self.module.init, seed_, *args, **kwargs)
-      return self.module.apply(state, *args, **kwargs)
+      if kwargs.get('mutable', False):
+        raise NotImplementedError
+      params = self._params(*args, **kwargs)
+      return self.module.apply(params, *args, **kwargs)
+
+    def _params(self, *args, **kwargs):
+      if self.treedef is None:
+        params = self.module.init(seed(), *args, **kwargs)
+        params = postinit(params) if postinit else params
+        mapping, self.treedef = flatten(params)
+        [self.value(k, v) for k, v in mapping.items()]
+      return unflatten(self.values, self.treedef)
+
   return FlaxModule
 
 
-def FromHaiku(ctor):
+def FromHaiku(make):
+
   class HaikuModule(Module):
+
     def __init__(self, *args, **kwargs):
       import haiku as hk
-      def net(*args_, **kwargs_):
-        return ctor(*args, **kwargs)(*args_, **kwargs_)
+      def net(*a, **k):
+        return make(*args, **kwargs)(*a, **k)
       self.transformed = hk.transform(net)
+      self.treedef = None
+
     def __call__(self, *args, **kwargs):
-      seed_ = seed() if creating() else None
-      state = self.get('haiku', self.transformed.init, seed_, *args, **kwargs)
-      return self.transformed.apply(state, seed_, *args, **kwargs)
+      params = self._params(*args, **kwargs)
+      return self.transformed.apply(params, seed(), *args, **kwargs)
+
+    def _params(self, *args, **kwargs):
+      if self.treedef is None:
+        params = self.transformed.init(seed(), *args, **kwargs)
+        mapping, self.treedef = flatten(params)
+        [self.value(k, v) for k, v in mapping.items()]
+      return unflatten(self.values, self.treedef)
+
   return HaikuModule
 
 
-def FromOptax(ctor):
+def FromOptax(make):
+
   class OptaxModule(Module):
+
     def __init__(self, *args, **kwargs):
-      self.opt = ctor(*args, **kwargs)
+      if callable(make):
+        self.opt = make(*args, **kwargs)
+      else:
+        assert not args and not kwargs
+        self.opt = make
+      self.treedef = None
+
     def __call__(self, loss, keys, *args, **kwargs):
       import optax
       loss, params, grads = grad(loss, keys)(*args, **kwargs)
-      optstate = self.get('state', self.opt.init, params)
-      updates, optstate = self.opt.update(grads, optstate)
-      self.put('state', optstate)
+      updates = self.update(grads, params)
       context().update(optax.apply_updates(params, updates))
       return loss, params, grads
+
+    def update(self, grads, params):
+      if self.treedef is None:
+        mapping, self.treedef = flatten(self.opt.init(params))
+        [self.value(k, v) for k, v in mapping.items()]
+      state = unflatten(self.values, self.treedef)
+      updates, state = self.opt.update(grads, state)
+      [self.write(k, v) for k, v in flatten(state)[0].items()]
+      return updates
+
   return OptaxModule
